use std::collections::HashMap;

use crate::{
    expr::Expr,
    stmt::Stmt,
    token::{Token, TokenType},
};

pub struct Parser {
    tokens: Vec<Token>,
    current: usize,
    env: Vec<char>,
}

impl Parser {
    pub fn new(tokens: Vec<Token>) -> Self {
        Self {
            tokens,
            current: 0,
            env: vec![],
        }
    }

    pub fn parse(&mut self) -> Option<(HashMap<String, Expr>, Expr)> {
        let mut definitions = HashMap::new();
        let mut main_expr = None;
        while !self.is_at_end() {
            let stmt = self.stmt()?;

            match stmt {
                Stmt::Definition(name, expr) => {
                    definitions.insert(name, expr);
                }

                Stmt::Main(expr) => {
                    if main_expr.is_some() {
                        self.print_error("Duplicate primary expression");
                        return None;
                    }

                    main_expr = Some(expr);
                }
            }
        }

        match main_expr {
            Some(expr) => Some((definitions, expr)),
            None => {
                self.print_error("Missing primary expression");

                None
            }
        }
    }

    fn stmt(&mut self) -> Option<Stmt> {
        let stmt = if *self.peek_next() == TokenType::Equals {
            // definition
            let TokenType::DefinitionName(name) = &self.advance().token_type else {
                self.print_error("Expected definition name before '='");
                return None;
            };

            let name = name.clone();

            // consume =
            self.advance();

            let expr = self.expr()?;

            Stmt::Definition(name, expr)
        } else {
            Stmt::Main(self.expr()?)
        };

        // consume semicolon
        if self.peek().token_type != TokenType::Semicolon {
            self.print_error("Expected ';' after statement");
            return None;
        }

        self.advance();

        Some(stmt)
    }

    fn expr(&mut self) -> Option<Expr> {
        let mut expr = self.expr_once()?;

        while !self.is_at_end_of_expr() {
            let right = self.expr_once()?;
            expr = Expr::App(Box::new(expr), Box::new(right));
        }

        Some(expr)
    }

    fn expr_once(&mut self) -> Option<Expr> {
        match &self.peek().token_type {
            TokenType::Lambda => self.abstraction(),

            TokenType::DefinitionName(name) => {
                let name = name.clone();
                self.advance();

                Some(Expr::DefinitionName(name))
            }

            TokenType::LeftParen => {
                self.advance(); // consume (

                let expr = self.expr()?;

                if self.peek().token_type != TokenType::RightParen {
                    self.print_error("Expected ')'");
                    return None;
                }

                self.advance(); // consume )

                Some(expr)
            }

            TokenType::Variable(c) => {
                let c = *c;
                self.advance();

                Some(Expr::Var(self.find_variable(c)?))
            }

            _ => {
                self.print_error(
                    "Unexpected token. Expected one of 'λ', '/', '\\', '$', '(', or a variable",
                );
                None
            }
        }
    }

    fn abstraction(&mut self) -> Option<Expr> {
        self.advance(); // consume lambda

        let mut abs_count = 0;
        while let TokenType::Variable(v) = self.peek().token_type {
            self.env.push(v);
            abs_count += 1;
            self.advance();
        }

        if self.advance().token_type != TokenType::Period {
            self.print_error("Expected '.' after abstraction variable definitions");
            return None;
        }

        let mut expr = self.expr()?;

        for _ in 0..abs_count {
            self.env.pop().unwrap();
            expr = Expr::Abs(Box::new(expr));
        }

        Some(expr)
    }

    fn find_variable(&self, name: char) -> Option<usize> {
        let ret = self
            .env
            .iter()
            .enumerate()
            .find(|(_, x)| **x == name)
            .map(|(i, _)| i);

        match ret {
            Some(x) => Some(self.env.len() - x - 1),
            None => {
                self.print_error(&format!("Unknown variable '{}'.", name));

                None
            }
        }
    }

    fn peek(&self) -> &Token {
        &self.tokens[self.current]
    }

    fn peek_next(&self) -> &TokenType {
        if self.current + 1 == self.tokens.len() {
            &TokenType::NoToken
        } else {
            &self.tokens[self.current + 1].token_type
        }
    }

    fn advance(&mut self) -> &Token {
        let t = &self.tokens[self.current];
        self.current += 1;

        t
    }

    fn is_at_end(&self) -> bool {
        self.peek().token_type == TokenType::Eof
    }

    fn is_at_end_of_expr(&self) -> bool {
        let tt = &self.peek().token_type;

        // end of expr can be ), $, or end of file
        matches!(tt, TokenType::RightParen | TokenType::Semicolon)
    }

    fn print_error(&self, err: &str) {
        eprintln!("Error on line [{}]: {err}", self.peek().line);
    }
}
